"""Client for performing workcell actions"""

import json
import time
from pathlib import Path
from typing import Any, Optional

import requests
from madsci.common.data_manipulation import value_substitution, walk_and_replace
from madsci.common.exceptions import WorkflowFailedError
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
    WorkflowStatus,
)


class WorkcellClient:
    """a client for running workflows"""

    def __init__(
        self,
        workcell_manager_url: str,
        working_directory: str = "~/.MADsci/temp",
        ownership_info: Optional[OwnershipInfo] = None,
    ) -> "WorkcellClient":
        """initialize the client"""
        self.url = workcell_manager_url
        self.working_directory = Path(working_directory).expanduser()
        self.ownership_info = ownership_info

    def query_workflow(self, workflow_id: str) -> Optional[Workflow]:
        """Checks on a workflow run using the id given

        Parameters
        ----------

        workflow_id : str
           The id returned by the start_workflow function for this run

        Returns
        -------

        response: Dict
           The JSON portion of the response from the server"""

        url = f"{self.url}/workflows/{workflow_id}"
        response = requests.get(url, timeout=10)

        if response.ok:
            return Workflow(**response.json())
        response.raise_for_status()
        return None

    def submit_workflow(
        self,
        workflow: str,
        parameters: dict,
        validate_only: bool = False,
        blocking: bool = True,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """Send a workflow to the workcell manager"""
        workflow = WorkflowDefinition.from_yaml(workflow)
        WorkflowDefinition.model_validate(workflow)
        insert_parameter_values(workflow=workflow, parameters=parameters)
        files = self._extract_files_from_workflow(workflow)
        url = self.url + "/workflows/start"
        response = requests.post(
            url,
            data={
                "workflow": workflow.model_dump_json(),
                "parameters": json.dumps(parameters),
                "validate_only": validate_only,
                "ownership_info": self.ownership_info.model_dump(mode="json")
                if self.ownership_info
                else None,
            },
            files={
                (
                    "files",
                    (
                        str(Path(path).name),
                        Path.open(Path(path).expanduser(), "rb"),
                    ),
                )
                for _, path in files.items()
            },
            timeout=10,
        )
        if not blocking:
            return Workflow(**response.json())
        return self.await_workflow(
            response.json()["workflow_id"],
            raise_on_cancelled=raise_on_cancelled,
            raise_on_failed=raise_on_failed,
        )

    def _extract_files_from_workflow(
        self, workflow: WorkflowDefinition
    ) -> dict[str, Any]:
        """
        Returns a dictionary of files from a workflow
        """
        files = {}
        for step in workflow.steps:
            if step.files:
                for file, path in step.files.items():
                    unique_filename = f"{step.step_id}_{file}"
                    files[unique_filename] = path
                    if not Path(files[unique_filename]).is_absolute():
                        files[unique_filename] = (
                            self.working_directory / files[unique_filename]
                        )
                    step.files[file] = Path(files[unique_filename]).name
        return files

    def submit_workflow_sequence(
        self, workflows: list[str], parameters: list[dict[str:Any]]
    ) -> list[Workflow]:
        """Submit a list of workflows to run in a specific order"""
        wfs = []
        for i in range(len(workflows)):
            wf = self.submit_workflow(workflows[i], parameters[i], blocking=True)
            wfs.append(wf)
        return wfs

    def submit_workflow_batch(
        self, workflows: list[str], parameters: list[dict[str:Any]]
    ) -> list[Workflow]:
        """Submit a batch of workflows to run in no particular order"""
        id_list = []
        for i in range(len(workflows)):
            response = self.submit_workflow(workflows[i], parameters[i], blocking=False)
            id_list.append(response.json()["workflow_id"])
        finished = False
        while not finished:
            flag = True
            wfs = []
            for id in id_list:
                wf = self.query_workflow(id)
                flag = flag and (wf.status in ["completed", "failed"])
                wfs.append(wf)
            finished = flag
        return wfs

    def retry_workflow(self, workflow_id: str, index: int = -1) -> dict:
        """rerun an exisiting wf using the same wf id"""
        url = f"{self.url}/workflows/retry"
        response = requests.post(
            url,
            params={
                "workflow_id": workflow_id,
                "index": index,
            },
            timeout=10,
        )
        return response.json()

    def resubmit_workflow(
        self,
        workflow_id: str,
        blocking: bool = True,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """resubmit an existing workflows as a new workflow with a new id"""
        url = f"{self.url}/workflows/resubmit/{workflow_id}"
        response = requests.get(url, timeout=10)
        new_wf = Workflow(**response.json())
        if blocking:
            return self.await_workflow(
                new_wf.workflow_id,
                raise_on_failed=raise_on_failed,
                raise_on_cancelled=raise_on_cancelled,
            )
        return new_wf

    def await_workflow(
        self,
        workflow_id: str,
        raise_on_failed: bool = True,
        raise_on_cancelled: bool = True,
    ) -> Workflow:
        """await a workflows completion"""
        prior_status = None
        prior_index = None
        while True:
            wf = self.query_workflow(workflow_id)
            status = wf.status
            step_index = wf.step_index
            if prior_status != status or prior_index != step_index:
                if step_index < len(wf.steps):
                    step_name = wf.steps[step_index].name
                else:
                    step_name = "Workflow End"
                # TODO: Improve progress reporting
                print(  # noqa: T201
                    f"\n{wf.name} [{step_index}]: {step_name} ({wf.status})",
                    end="",
                    flush=True,
                )
            else:
                print(".", end="", flush=True)  # noqa: T201
            time.sleep(1)
            if wf.status in [
                WorkflowStatus.COMPLETED,
                WorkflowStatus.FAILED,
                WorkflowStatus.CANCELLED,
            ]:
                break
            prior_status = status
            prior_index = step_index
        if wf.status == WorkflowStatus.FAILED and raise_on_failed:
            raise WorkflowFailedError(
                f"Workflow {wf.name} ({wf.workflow_id}) failed on step {wf.step_index}: '{wf.steps[wf.step_index].name}'."
            )
        if wf.status == WorkflowStatus.CANCELLED and raise_on_cancelled:
            raise WorkflowFailedError(
                f"Workflow {wf.name} ({wf.workflow_id}) was cancelled on step {wf.step_index}: '{wf.steps[wf.step_index].name}'."
            )
        return wf

    def get_all_nodes(self) -> dict:
        """get all nodes in the workcell"""
        url = f"{self.url}/nodes"
        response = requests.get(url, timeout=10)
        return response.json()

    def get_node(self, node_name: str) -> dict:
        """get a single node from a workcell"""
        url = f"{self.url}/nodes/{node_name}"
        response = requests.get(url, timeout=10)
        return response.json()

    def add_node(
        self,
        node_name: str,
        node_url: str,
        node_description: str = "A Node",
        permanent: bool = False,
    ) -> dict:
        """add a node to a workcell"""
        url = f"{self.url}/nodes/add_node"
        response = requests.post(
            url,
            params={
                "node_name": node_name,
                "node_url": node_url,
                "node_description": node_description,
                "permanent": permanent,
            },
            timeout=10,
        )
        return response.json()

    def get_all_workflows(self) -> dict:
        """get all workflows from a workcell manager"""
        url = f"{self.url}/workflows"
        response = requests.get(url, timeout=100)
        return response.json()

    def get_workcell_state(self) -> dict:
        """Get the full state of the workcell"""
        url = f"{self.url}/state"
        response = requests.get(url, timeout=10)
        return response.json()


def insert_parameter_values(
    workflow: WorkflowDefinition, parameters: dict[str, Any]
) -> Workflow:
    """Replace the parameter strings in the workflow with the provided values"""
    for param in workflow.parameters:
        if param.name not in parameters:
            if param.default:
                parameters[param.name] = param.default
            else:
                raise ValueError(
                    "Workflow parameter: "
                    + param.name
                    + " not provided, and no default value is defined."
                )
    steps = []
    for step in workflow.steps:
        for key, val in iter(step):
            if type(val) is str:
                setattr(step, key, value_substitution(val, parameters))

        step.args = walk_and_replace(step.args, parameters)
        steps.append(step)
    workflow.steps = steps



"""Client for the MADSci Experiment Manager."""

from typing import Optional, Union

import requests
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.experiment_types import (
    Experiment,
    ExperimentalCampaign,
    ExperimentDesign,
    ExperimentRegistration,
    ExperimentStatus,
)
from pydantic import AnyUrl
from ulid import ULID


class ExperimentClient:
    """Client for the MADSci Experiment Manager."""

    url: AnyUrl

    def __init__(
        self, url: Union[str, AnyUrl], ownership_info: Optional[OwnershipInfo] = None
    ) -> "ExperimentClient":
        """Create a new Experiment Client."""
        self.url = AnyUrl(url)
        self.ownership_info = ownership_info if ownership_info else OwnershipInfo()

    def get_experiment(self, experiment_id: Union[str, ULID]) -> dict:
        """Get an experiment by ID."""
        response = requests.get(f"{self.url}/experiment/{experiment_id}", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def get_experiments(self, number: int = 10) -> list[Experiment]:
        """Get a list of the latest experiments."""
        response = requests.get(
            f"{self.url}/experiments", params={number: number}, timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return [Experiment.model_validate(experiment) for experiment in response.json()]

    def start_experiment(
        self,
        experiment_design: ExperimentDesign,
        run_name: Optional[str] = None,
        run_description: Optional[str] = None,
    ) -> Experiment:
        """Start an experiment based on an ExperimentDesign."""
        response = requests.post(
            f"{self.url}/experiment",
            json=ExperimentRegistration(
                experiment_design=experiment_design.model_dump(mode="json"),
                run_name=run_name,
                run_description=run_description,
            ).model_dump(mode="json"),
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def end_experiment(
        self, experiment_id: Union[str, ULID], status: Optional[ExperimentStatus] = None
    ) -> Experiment:
        """End an experiment by ID. Optionally, set the status."""
        response = requests.post(
            f"{self.url}/experiment/{experiment_id}/end",
            params={"status": status},
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def continue_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Continue an experiment by ID."""
        response = requests.post(
            f"{self.url}/experiment/{experiment_id}/continue", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def pause_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Pause an experiment by ID."""
        response = requests.post(
            f"{self.url}/experiment/{experiment_id}/pause", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def cancel_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Cancel an experiment by ID."""
        response = requests.post(
            f"{self.url}/experiment/{experiment_id}/cancel", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def register_campaign(self, campaign: ExperimentalCampaign) -> ExperimentalCampaign:
        """Register a new experimental campaign."""
        response = requests.post(f"{self.url}/campaign", json=campaign, timeout=10)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def get_campaign(self, campaign_id: str) -> ExperimentalCampaign:
        """Get an experimental campaign by ID."""
        response = requests.get(f"{self.url}/campaign/{campaign_id}", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return response.json()



"""CLI for interacting with resources."""

from pathlib import Path
from typing import Optional

import click
from click.core import Context
from madsci.common.types.resource_types import (
    RESOURCE_TYPE_MAP,
)
from madsci.common.types.resource_types.custom_types import (
    ResourceTypeEnum,
)
from madsci.common.types.resource_types.definitions import (
    ResourceDefinition,
    ResourceFile,
)
from madsci.common.utils import (
    prompt_for_input,
    prompt_from_list,
    prompt_from_pydantic_model,
    prompt_yes_no,
    save_model,
    search_for_file_pattern,
)
from rich.console import Console
from rich.pretty import pprint

console = Console()


class ResourceContext:
    """Context object for resource commands."""

    def __init__(self) -> None:
        """Initialize the context object."""
        self.resource_file: Optional[ResourceFile] = None
        self.path: Optional[Path] = None


pass_resource = click.make_pass_decorator(ResourceContext)


def find_resource_file(path: Optional[str]) -> ResourceContext:
    """Find a resource file by path."""
    resource_context = ResourceContext()

    if path:
        resource_context.path = Path(path)
        if resource_context.path.exists():
            resource_context.resource_file = ResourceFile.from_yaml(path)
            return resource_context

    # Search for any resource file
    resource_files = search_for_file_pattern("*.resources.yaml")
    if resource_files:
        resource_context.path = Path(resource_files[0])
        resource_context.resource_file = ResourceFile.from_yaml(resource_files[0])

    return resource_context


@click.group()
@click.option("--path", "-p", type=str, help="Path to the resource definition file.")
@click.pass_context
def resource(ctx: Context, path: Optional[str]) -> None:
    """Manage resources."""
    ctx.obj = find_resource_file(path)


@resource.command()
@click.pass_context
def create(ctx: Context) -> None:
    """Create a new resource file."""
    path = ctx.parent.params.get("path")
    if not path:
        default_path = Path.cwd() / "default.resources.yaml"
        new_path = prompt_for_input(
            "Path to save Resource Definition file",
            default=str(default_path),
        )
        if new_path:
            path = Path(new_path)

    resource_file = ResourceFile()
    save_model(path=path, model=resource_file)
    console.print(f"Created resource file: {path}")


@resource.command()
def list() -> None:
    """List all resource files and their contents."""
    resource_files = search_for_file_pattern("*.resources.yaml")

    if resource_files:
        for resource_file in sorted(set(resource_files)):
            resource_def = ResourceFile.from_yaml(resource_file)
            console.print(f"\n[bold]Resource File[/]: {resource_file}")

            if resource_def.resource_types:
                console.print("\n[bold]Resource Types:[/]")
                for resource_type in resource_def.resource_types:
                    console.print(
                        f"  [bold]{resource_type.type_name}[/]: {resource_type.type_description}",
                    )

            if resource_def.default_resources:
                console.print("\n[bold]Default Resources:[/]")
                for resource in resource_def.default_resources:
                    console.print(
                        f"  [bold]{resource.resource_name}[/]: {resource.resource_description or 'No description'}",
                    )
    else:
        console.print("No resource files found")


@resource.group(name="type")
def resource_type() -> None:
    """Manage resource types within a resource file."""


@resource_type.command()
@click.option("--name", "-n", type=str, help="Name of the resource type.")
@click.option("--description", "-d", type=str, help="Description of the resource type.")
@click.option("--base-type", "-b", type=str, help="Base type of the resource.")
@pass_resource
def add(
    ctx: ResourceContext,
    name: Optional[str],
    description: Optional[str],
    base_type: Optional[str],
) -> None:
    """Add a new resource type to the resource file."""
    if not ctx.resource_file or not ctx.path:
        console.print(
            "No resource file found. Create one with 'madsci resource create' first.",
        )
        return

    if not name:
        name = prompt_for_input("Resource Type Name", required=True)
    if not description:
        description = prompt_for_input("Resource Type Description")
    if not base_type or base_type not in ResourceTypeEnum:
        base_type = prompt_from_list(
            "Base Type",
            [e.value for e in ResourceTypeEnum],
            default=ResourceTypeEnum.resource.value,
        )
    # *Get the appropriate type definition class
    type_def_class = RESOURCE_TYPE_MAP[base_type]["definition"]

    # *Create the type definition with the fields we've collected
    type_def = type_def_class(
        **prompt_from_pydantic_model(
            type_def_class,
            "Resource Type Definition",
            type_name=name,
            type_description=description,
            base_type=base_type,
        ),
    )

    # *Check if type already exists
    if any(rt.type_name == name for rt in ctx.resource_file.resource_types):
        if not prompt_yes_no(
            f"Resource type '{name}' already exists. Overwrite?",
            default=False,
        ):
            return
        # *Remove existing type
        ctx.resource_file.resource_types = [
            rt for rt in ctx.resource_file.resource_types if rt.type_name != name
        ]

    # *Add the type definition to the resource file
    ctx.resource_file.resource_types.append(type_def)
    save_model(ctx.path, ctx.resource_file, overwrite_check=False)
    console.print(f"Added resource type: [bold]{name}[/]")


@resource_type.command()
@click.argument("name", required=False)
@pass_resource
def delete(ctx: ResourceContext, name: Optional[str]) -> None:
    """Delete a resource type from the resource file."""
    if not ctx.resource_file or not ctx.path:
        console.print(
            "No resource file found. Create one with 'madsci resource create' first.",
        )
        return

    if not ctx.resource_file.resource_types:
        console.print("No resource types defined in this file.")
        return

    if not name:
        name = prompt_from_list(
            "Resource Type to Delete",
            [rt.type_name for rt in ctx.resource_file.resource_types],
            required=True,
        )

    # Find the resource type
    resource_type = next(
        (rt for rt in ctx.resource_file.resource_types if rt.type_name == name),
        None,
    )
    if not resource_type:
        console.print(f"Resource type [bold]{name}[/] not found.")
        return

    # Check if type is used by any default resources
    used_by_resources = [
        r.resource_name
        for r in ctx.resource_file.default_resources
        if r.resource_type == name
    ]
    if used_by_resources:
        console.print(
            f"Cannot delete resource type [bold]{name}[/] as it is used by these resources:",
        )
        for resource_name in used_by_resources:
            console.print(f"  - {resource_name}")
        return

    if prompt_yes_no(f"Delete resource type [bold]{name}[/]?", default=False):
        ctx.resource_file.resource_types = [
            rt for rt in ctx.resource_file.resource_types if rt.type_name != name
        ]
        save_model(ctx.path, ctx.resource_file, overwrite_check=False)
        console.print(f"Deleted resource type: [bold]{name}[/]")


@resource_type.command()
@click.argument("name", required=False)
@pass_resource
def info(ctx: ResourceContext, name: Optional[str]) -> None:
    """Show information about a resource type."""
    if not ctx.resource_file:
        console.print(
            "No resource file found. Create one with 'madsci resource create' first.",
        )
        return

    if not ctx.resource_file.resource_types:
        console.print("No resource types defined in this file.")
        return

    if not name:
        name = prompt_from_list(
            "Resource Type",
            [rt.type_name for rt in ctx.resource_file.resource_types],
            required=True,
        )

    resource_type = next(
        (rt for rt in ctx.resource_file.resource_types if rt.type_name == name),
        None,
    )
    if resource_type:
        pprint(resource_type)
    else:
        console.print(f"Resource type [bold]{name}[/] not found.")


@resource_type.command(name="list")
@pass_resource
def list_types(ctx: ResourceContext) -> None:
    """List all resource types in the file."""
    if not ctx.resource_file:
        console.print(
            "No resource file found. Create one with 'madsci resource create' first.",
        )
        return

    if not ctx.resource_file.resource_types:
        console.print("No resource types defined in this file.")
        return

    console.print("\n[bold]Resource Types:[/]")
    for resource_type in ctx.resource_file.resource_types:
        console.print(
            f"  [bold]{resource_type.type_name}[/] ({resource_type.base_type})",
        )
        if resource_type.type_description:
            console.print(f"    Description: {resource_type.type_description}")
        console.print(f"    Parent Types: {', '.join(resource_type.parent_types)}")


@resource.command()
@pass_resource
def add_resource(ctx: ResourceContext) -> None:
    """Add a new default resource to the resource file."""
    if not ctx.resource_file or not ctx.path:
        console.print(
            "No resource file found. Create one with 'madsci resource create' first.",
        )
        return

    name = prompt_for_input("Resource Name", required=True)
    description = prompt_for_input("Resource Description")

    # * Combine built-in types and custom types for selection
    available_types = [e.value for e in ResourceTypeEnum]
    custom_types = [rt.type_name for rt in ctx.resource_file.resource_types]
    all_types = available_types + custom_types

    resource_type = prompt_from_list(
        "Resource Type",
        all_types,
        default=ResourceTypeEnum.resource.value,
    )

    # * Create the resource definition
    if resource_type in RESOURCE_TYPE_MAP:
        resource_def_class = RESOURCE_TYPE_MAP[resource_type]["definition"]
    else:
        resource_def_class = ResourceDefinition

    resource_def = resource_def_class(
        resource_name=name,
        resource_description=description,
        resource_type=resource_type,
    )

    ctx.resource_file.default_resources.append(resource_def)
    save_model(ctx.path, ctx.resource_file, overwrite_check=False)
    console.print(f"Added default resource: [bold]{name}[/]")


@resource.command(name="info")
@pass_resource
def file_info(ctx: ResourceContext) -> None:
    """Get information about a resource file."""
    if ctx.resource_file:
        pprint(ctx.resource_file)
    else:
        console.print(
            "No resource file found. Create one with 'madsci resource create'.",
        )


@resource.command()
@pass_resource
def validate(ctx: ResourceContext) -> None:
    """Validate a resource file."""
    if ctx.resource_file:
        console.print(ctx.resource_file)
    else:
        console.print(
            "No resource file found. Create one with 'madsci resource create'.",
        )


@resource.command(name="delete")
@pass_resource
def delete_file(ctx: ResourceContext) -> None:
    """Delete a resource file."""
    if ctx.resource_file and ctx.path:
        console.print(f"Deleting resource file: {ctx.path}")
        if prompt_yes_no("Are you sure?"):
            ctx.path.unlink()
            console.print(f"Deleted {ctx.path}")
    else:
        console.print(
            "No resource file found. Create one with 'madsci resource create'.",
        )



"""Command Line Interface for managing MADSci Squid labs."""

import os
import shlex
from pathlib import Path
from typing import Optional

import click
from click.core import Context
from madsci.common.types.lab_types import LabDefinition
from madsci.common.utils import (
    prompt_for_input,
    prompt_yes_no,
    save_model,
    search_for_file_pattern,
    to_snake_case,
)
from rich import print
from rich.console import Console
from rich.pretty import pprint

console = Console()


class LabContext:
    """Context object for lab commands."""

    def __init__(self) -> None:
        """Initialize the context object."""
        self.lab_def: Optional[LabDefinition] = None
        self.path: Optional[Path] = None
        self.quiet: bool = False


pass_lab = click.make_pass_decorator(LabContext)


def find_lab(name: Optional[str], path: Optional[str]) -> LabContext:
    """Find a lab by name or path."""
    lab_context = LabContext()

    if path:
        lab_context.path = Path(path)
        if lab_context.path.exists():
            lab_context.lab_def = LabDefinition.from_yaml(path)
            return lab_context

    if name:
        lab_files = search_for_file_pattern("*.lab.yaml")
        for lab_file in lab_files:
            lab_def = LabDefinition.from_yaml(lab_file)
            if lab_def.name == name:
                lab_context.path = Path(lab_file)
                lab_context.lab_def = lab_def
                return lab_context

    # * Search for any lab file
    lab_files = search_for_file_pattern("*.lab.yaml")
    if lab_files:
        lab_context.path = Path(lab_files[0])
        lab_context.lab_def = LabDefinition.from_yaml(lab_files[0])

    return lab_context


@click.group()
@click.option("--name", "-n", type=str, help="The name of the lab to operate on.")
@click.option("--path", "-p", type=str, help="The path to the lab definition file.")
@click.pass_context
def lab(ctx: Context, name: Optional[str], path: Optional[str]) -> None:
    """Manage labs."""
    ctx.obj = find_lab(name, path)
    ctx.obj.quiet = ctx.parent.params.get("quiet")


@lab.command()
@click.option("--name", "-n", type=str, help="The name of the lab.", required=False)
@click.option("--path", "-p", type=str, help="The path to the lab definition file.")
@click.option("--description", "-d", type=str, help="The description of the lab.")
@click.pass_context
def create(
    ctx: Context,
    name: Optional[str],
    path: Optional[str],
    description: Optional[str],
) -> None:
    """Create a new lab."""
    if not name:
        name = ctx.parent.params.get("name")
    if not name:
        name = prompt_for_input("Lab Name", required=True, quiet=ctx.obj.quiet)
    if not description:
        description = prompt_for_input("Lab Description", quiet=ctx.obj.quiet)

    lab_definition = LabDefinition(name=name, description=description)
    console.print(lab_definition)

    if not path:
        path = ctx.parent.params.get("path")
    if not path:
        default_path = Path.cwd() / f"{to_snake_case(name)}.lab.yaml"
        new_path = prompt_for_input(
            "Path to save Lab Definition file",
            default=str(default_path),
            quiet=ctx.obj.quiet,
        )
        if new_path:
            path = Path(new_path)
    print("Path:", path)
    save_model(path=path, model=lab_definition, overwrite_check=not ctx.obj.quiet)


@lab.command()
def list() -> None:
    """List all labs. Will list all labs in the current directory, subdirectories, and parent directories."""
    lab_files = search_for_file_pattern("*.lab.yaml")

    if lab_files:
        for lab_file in sorted(set(lab_files)):
            lab_definition = LabDefinition.from_yaml(lab_file)
            console.print(
                f"[bold]{lab_definition.name}[/]: {lab_definition.description} ({lab_file})",
            )
    else:
        print("No lab definitions found")


@lab.command()
@pass_lab
def info(ctx: LabContext) -> None:
    """Get information about a lab."""
    if ctx.lab_def:
        pprint(ctx.lab_def)
    else:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab file, you can create one with 'madsci lab create'.",
        )


@lab.command()
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@pass_lab
def delete(ctx: LabContext, yes: bool) -> None:
    """Delete a lab."""
    if ctx.lab_def and ctx.path:
        console.print(f"Deleting lab: {ctx.lab_def.name} ({ctx.path})")
        if yes or ctx.quiet or prompt_yes_no("Are you sure?"):
            ctx.path.unlink()
            console.print(f"Deleted {ctx.path}")
    else:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab file, you can create one with 'madsci lab create'.",
        )


@lab.command()
@pass_lab
def validate(ctx: LabContext) -> None:
    """Validate a lab definition file."""
    if ctx.lab_def:
        console.print(ctx.lab_def)
    else:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab definition file, you can create one with 'madsci lab create'.",
        )


def run_command(command: str, lab: LabDefinition, path: Path) -> None:
    """Run a command in a lab."""
    console.print(
        f"Running command: [bold]{command}[/] ({lab.commands[command]}) in lab: [bold]{lab.name}[/] ({path})",
    )
    args = shlex.split(lab.commands[command])
    os.execvp(args[0], args)  # noqa: S606


@lab.command()
@click.argument("command", type=str)
@pass_lab
def run(ctx: LabContext, command: str) -> None:
    """Run a command in a lab."""
    if not ctx.lab_def:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab file, you can create one with 'madsci lab create'.",
        )
        return

    if ctx.lab_def.commands.get(command):
        run_command(command, ctx.lab_def, ctx.path)
    else:
        console.print(
            f"Command [bold]{command}[/] not found in lab definition: [bold]{ctx.lab_def.name}[/] ({ctx.path})",
        )


@lab.command()
@click.option("--command_name", "--name", "-n", type=str, required=False)
@click.option("--command", "-c", type=str, required=False)
@pass_lab
def add_command(ctx: LabContext, command_name: str, command: str) -> None:
    """Add a command to a lab definition."""
    if not ctx.lab_def:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab file, you can create one with 'madsci lab create'.",
        )
        return

    if not command_name:
        command_name = prompt_for_input("Command Name", required=True)
    if not command:
        command = prompt_for_input("Command", required=True)

    if command_name in ctx.lab_def.commands:
        console.print(
            f"Command [bold]{command_name}[/] already exists in lab definition: [bold]{ctx.lab_def.name}[/] ({ctx.path})",
        )
        if not prompt_yes_no("Do you want to overwrite it?", default="no"):
            return

    ctx.lab_def.commands[command_name] = command
    save_model(ctx.path, ctx.lab_def, overwrite_check=False)
    console.print(
        f"Added command [bold]{command_name}[/] to lab: [bold]{ctx.lab_def.name}[/]",
    )


@lab.command()
@click.argument("command_name", type=str, required=False)
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@pass_lab
def delete_command(ctx: LabContext, command_name: str, yes: bool) -> None:
    """Delete a command from a lab definition."""
    if not ctx.lab_def:
        console.print(
            "No lab found. Specify lab by name or path. If you don't have a lab file, you can create one with 'madsci lab create'.",
        )
        return

    if not command_name:
        command_name = prompt_for_input("Command Name", required=True)

    if command_name in ctx.lab_def.commands:
        if (
            yes
            or ctx.quiet
            or prompt_yes_no(
                f"Are you sure you want to delete command [bold]{command_name}[/]?",
                default="no",
            )
        ):
            del ctx.lab_def.commands[command_name]
            save_model(ctx.path, ctx.lab_def, overwrite_check=False)
            console.print(
                f"Deleted command [bold]{command_name}[/] from lab: [bold]{ctx.lab_def.name}[/]",
            )
    else:
        console.print(
            f"Command [bold]{command_name}[/] not found in lab definition: [bold]{ctx.lab_def.name}[/] ({ctx.path})",
        )



"""Client for the MADSci Experiment Manager."""

import warnings
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Optional, Union

import requests
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.datapoint_types import (
    DataPoint,
)
from pydantic import AnyUrl
from ulid import ULID


class DataClient:
    """Client for the MADSci Experiment Manager."""

    url: AnyUrl

    def __init__(
        self,
        url: Optional[Union[str, AnyUrl]] = None,
        ownership_info: Optional[OwnershipInfo] = None,
    ) -> "DataClient":
        """Create a new Datapoint Client."""
        self.url = AnyUrl(url) if url is not None else None
        if self.url is None:
            warnings.warn(
                "No URL provided for the data client. Cannot persist datapoints.",
                UserWarning,
                stacklevel=2,
            )
        self._local_datapoints = {}
        self.ownership_info = ownership_info if ownership_info else OwnershipInfo()

    def get_datapoint(self, datapoint_id: Union[str, ULID]) -> dict:
        """Get an datapoint by ID."""
        if self.url is None:
            return self._local_datapoints[datapoint_id]
        response = requests.get(f"{self.url}datapoint/{datapoint_id}", timeout=10)
        response.raise_for_status()
        return DataPoint.discriminate(response.json())

    def get_datapoint_value(self, datapoint_id: Union[str, ULID]) -> Any:
        """Get an datapoint value by ID. If the datapoint is JSON, returns the JSON data.
        Otherwise, returns the raw data as bytes."""
        if self.url is None:
            if hasattr(self._local_datapoints[datapoint_id], "value"):
                return self._local_datapoints[datapoint_id].value
            if hasattr(self._local_datapoints[datapoint_id], "path"):
                with (
                    Path(self._local_datapoints[datapoint_id].path)
                    .resolve()
                    .expanduser()
                    .open("rb") as f
                ):
                    return f.read()
        response = requests.get(f"{self.url}datapoint/{datapoint_id}/value", timeout=10)
        response.raise_for_status()
        try:
            return response.json()
        except JSONDecodeError:
            return response.content

    def save_datapoint_value(
        self, datapoint_id: Union[str, ULID], output_filepath: str
    ) -> None:
        """Get an datapoint value by ID."""
        output_filepath = Path(output_filepath).expanduser()
        if self.url is None:
            if self._local_datapoints[datapoint_id].data_type == "file":
                import shutil

                shutil.copyfile(
                    self._local_datapoints[datapoint_id].path, output_filepath
                )
            else:
                with Path(output_filepath).open("w") as f:
                    f.write(str(self._local_datapoints[datapoint_id].value))
            return
        response = requests.get(f"{self.url}datapoint/{datapoint_id}/value", timeout=10)
        response.raise_for_status()
        try:
            with Path(output_filepath).open("w") as f:
                f.write(str(response.json()["value"]))

        except Exception:
            Path(output_filepath).expanduser().parent.mkdir(parents=True, exist_ok=True)
            with Path.open(output_filepath, "wb") as f:
                f.write(response.content)

    def get_datapoints(self, number: int = 10) -> list[DataPoint]:
        """Get a list of the latest datapoints."""
        if self.url is None:
            return list(self._local_datapoints.values()).sort(
                key=lambda x: x.datapoint_id, reverse=True
            )[:number]
        response = requests.get(
            f"{self.url}datapoints", params={number: number}, timeout=10
        )
        response.raise_for_status()
        return [DataPoint.discriminate(datapoint) for datapoint in response.json()]

    def submit_datapoint(self, datapoint: DataPoint) -> DataPoint:
        """Submit a Datapoint object"""
        if self.url is None:
            self._local_datapoints[datapoint.datapoint_id] = datapoint
            return datapoint
        if datapoint.data_type == "file":
            files = {
                (
                    "files",
                    (
                        str(Path(datapoint.path).name),
                        Path.open(Path(datapoint.path).expanduser(), "rb"),
                    ),
                )
            }
        else:
            files = {}
        response = requests.post(
            "/datapoint",
            data={"datapoint": datapoint.model_dump_json()},
            files=files,
            timeout=10,
        )
        response.raise_for_status()
        return DataPoint.discriminate(response.json())

    def query_datapoints(self, selector: Any) -> dict[str, DataPoint]:
        """Query datapoints based on a selector."""
        if self.url is None:
            return {
                datapoint_id: datapoint
                for datapoint_id, datapoint in self._local_datapoints.items()
                if selector(datapoint)
            }
        response = requests.post(
            f"{self.url}datapoints/query", json=selector, timeout=10
        )
        response.raise_for_status()
        return {
            datapoint_id: DataPoint.discriminate(datapoint)
            for datapoint_id, datapoint in response.json().items()
        }



"""Command Line Interface for managing MADSci Squid managers."""

from pathlib import Path
from typing import Optional

import click
from click.core import Context
from madsci.common.types.datapoint_types import DataManagerDefinition
from madsci.common.types.event_types import EventManagerDefinition
from madsci.common.types.experiment_types import ExperimentManagerDefinition
from madsci.common.types.lab_types import (
    LabDefinition,
    ManagerDefinition,
    ManagerType,
)
from madsci.common.types.resource_types.definitions import ResourceManagerDefinition
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.utils import (
    prompt_for_input,
    prompt_from_list,
    prompt_yes_no,
    save_model,
    search_for_file_pattern,
    to_snake_case,
)
from rich.console import Console
from rich.pretty import pprint

console = Console()


class ManagerContext:
    """Context object for manager commands."""

    def __init__(self) -> None:
        """Initialize the context object."""
        self.manager_def: Optional[ManagerDefinition] = None
        self.lab_def: Optional[LabDefinition] = None
        self.path: Optional[Path] = None
        self.quiet: bool = False


pass_manager = click.make_pass_decorator(ManagerContext)


def find_manager(name: Optional[str], path: Optional[str]) -> ManagerContext:
    """Find a manager by name or path."""
    manager_context = ManagerContext()

    if path:
        manager_context.path = Path(path)
        if manager_context.path.exists():
            manager_context.manager_def = ManagerDefinition.from_yaml(path)
            return manager_context

    if name:
        lab_files = search_for_file_pattern("*.lab.yaml")
        for lab_file in lab_files:
            lab_def = LabDefinition.from_yaml(lab_file)
            for manager_name, manager in lab_def.managers.items():
                if manager_name == name:
                    manager_context.path = Path(lab_file)
                    manager_context.lab_def = lab_def
                    manager_context.manager_def = (
                        manager
                        if isinstance(manager, ManagerDefinition)
                        else ManagerDefinition.from_yaml(manager)
                    )
                    return manager_context

    # * Search for any manager file
    manager_files = search_for_file_pattern("*.manager.yaml")
    if manager_files:
        manager_context.path = Path(manager_files[0])
        manager_context.manager_def = ManagerDefinition.from_yaml(manager_files[0])

    return manager_context


@click.group()
@click.option("--name", "-n", type=str, help="The name of the manager to operate on.")
@click.option("--path", "-p", type=str, help="The path to the manager definition file.")
@click.pass_context
def manager(ctx: Context, name: Optional[str], path: Optional[str]) -> None:
    """Manage lab system managers."""
    ctx.obj = find_manager(name, path)
    ctx.obj.quiet = ctx.parent.params.get("quiet")


@manager.command(name="add")
@click.option("--name", "-n", type=str, help="The name of the manager.", required=False)
@click.option("--path", "-p", type=str, help="The path to the manager definition file.")
@click.option("--description", "-d", type=str, help="The description of the manager.")
@click.option(
    "--manager_type",
    "-t",
    type=click.Choice([e.value for e in ManagerType]),
    help="The type of the manager.",
)
@click.pass_context
def add(
    ctx: Context,
    name: Optional[str],
    path: Optional[str],
    description: Optional[str],
    manager_type: str,
) -> None:
    """Add a new manager."""
    name = name if name else ctx.parent.params.get("name")
    name = (
        name
        if name
        else prompt_for_input("Manager Name", required=True, quiet=ctx.obj.quiet)
    )
    description = (
        description
        if description
        else prompt_for_input("Manager Description", quiet=ctx.obj.quiet)
    )
    manager_type = (
        manager_type
        if manager_type
        else prompt_from_list(
            "Manager Type",
            options=[e.value for e in ManagerType],
            quiet=ctx.obj.quiet,
            required=True,
        )
    )

    manager_definition = ManagerDefinition(
        name=name, description=description, manager_type=manager_type
    )
    console.print(manager_definition)

    if ctx.obj.lab_def:
        ctx.obj.lab_def.managers[name] = manager_definition
        save_model(ctx.obj.path, ctx.obj.lab_def, overwrite_check=not ctx.obj.quiet)
    else:
        if not path:
            path = ctx.parent.params.get("path")
        if not path:
            if ctx.obj.lab_def and ctx.obj.lab_def._definition_path:
                working_path = Path(ctx.obj.lab_def._definition_path).parent
            else:
                working_path = Path.cwd()
            if working_path.parts[-1] != "managers":
                working_path = working_path / "managers"
            default_path = working_path / f"{to_snake_case(name)}.manager.yaml"
            new_path = prompt_for_input(
                "Path to save Manager Definition file",
                default=str(default_path),
                quiet=ctx.obj.quiet,
            )
            if new_path:
                path = Path(new_path)
                path.expanduser().parent.mkdir(parents=True, exist_ok=True)
        manager_definition = promote_manager_definition(manager_definition)
        save_model(
            path=path, model=manager_definition, overwrite_check=not ctx.obj.quiet
        )


def promote_manager_definition(
    manager_definition: ManagerDefinition,
) -> ManagerDefinition:
    """Promote a manager definition to a more specific type."""
    if manager_definition.manager_type == ManagerType.EXPERIMENT_MANAGER:
        return ExperimentManagerDefinition(**manager_definition.model_dump(mode="json"))
    if manager_definition.manager_type == ManagerType.WORKCELL_MANAGER:
        return WorkcellDefinition(**manager_definition.model_dump(mode="json"))
    if manager_definition.manager_type == ManagerType.EVENT_MANAGER:
        return EventManagerDefinition(**manager_definition.model_dump(mode="json"))
    if manager_definition.manager_type == ManagerType.RESOURCE_MANAGER:
        return ResourceManagerDefinition(**manager_definition.model_dump(mode="json"))
    if manager_definition.manager_type == ManagerType.DATA_MANAGER:
        return DataManagerDefinition(**manager_definition.model_dump(mode="json"))
    return manager_definition


@manager.command()
def list() -> None:
    """List all managers. Will list all managers in the current directory, subdirectories, and parent directories."""
    manager_files = search_for_file_pattern("*.manager.yaml")

    if manager_files:
        for manager_file in sorted(set(manager_files)):
            manager_definition = ManagerDefinition.from_yaml(manager_file)
            console.print(
                f"[bold]{manager_definition.name}[/]: {manager_definition.description} ({manager_file})",
            )
    else:
        lab_files = search_for_file_pattern("*.lab.yaml")
        for lab_file in lab_files:
            lab_def = LabDefinition.from_yaml(lab_file)
            for _, manager in lab_def.managers.items():
                manager_definition = (
                    manager
                    if isinstance(manager, ManagerDefinition)
                    else ManagerDefinition.from_yaml(manager)
                )
                console.print(
                    f"[bold]{manager_definition.name}[/]: {manager_definition.description} (defined in {lab_file})",
                )


@manager.command()
@pass_manager
def info(ctx: ManagerContext) -> None:
    """Get information about a manager."""
    if ctx.manager_def:
        pprint(ctx.manager_def)
    else:
        console.print(
            "No manager found. Specify manager by name or path. If you don't have a manager file, you can create one with 'madsci manager create'.",
        )


@manager.command()
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt.")
@pass_manager
def delete(ctx: ManagerContext, yes: bool) -> None:
    """Delete a manager."""
    if ctx.manager_def and ctx.path:
        console.print(f"Deleting manager: {ctx.manager_def.name} ({ctx.path})")
        if yes or ctx.quiet or prompt_yes_no("Are you sure?"):
            if ctx.lab_def and ctx.lab_def.managers.get(ctx.manager_def.name):
                del ctx.lab_def.managers[ctx.manager_def.name]
                save_model(
                    ctx.obj.path, ctx.obj.lab_def, overwrite_check=not ctx.obj.quiet
                )
            else:
                ctx.path.unlink()
            console.print(f"Deleted {ctx.path}")
    else:
        console.print(
            "No manager found. Specify manager by name or path. If you don't have a manager file, you can create one with 'madsci manager create'.",
        )


@manager.command()
@pass_manager
def validate(ctx: ManagerContext) -> None:
    """Validate a manager definition file."""
    if ctx.manager_def:
        console.print(ctx.manager_def)
    else:
        console.print(
            "No manager found. Specify manager by name or path. If you don't have a manager definition file, you can create one with 'madsci manager create'.",
        )



"""Fast API Client for Resources"""

from datetime import datetime
from typing import Any, Optional, Union

import requests
from madsci.common.types.resource_types import (
    GridIndex2D,
    GridIndex3D,
    ResourceDataModels,
)
from madsci.common.types.resource_types.server_types import (
    PushResourceBody,
    RemoveChildBody,
    ResourceGetQuery,
    ResourceHistoryGetQuery,
    SetChildBody,
)
from madsci.resource_manager.resource_tables import Resource


class ResourceClient:
    """Fast API Client for Resources"""

    def __init__(self, url: str) -> None:
        """Initial params"""
        self.url = url
        if self.url.endswith("/"):
            self.url = {self.url[:-1]}

    def add_resource(self, resource: ResourceDataModels) -> ResourceDataModels:
        """
        Add a resource to the server.

        Args:
            resource (ResourceDataModels): The resource to add.

        Returns:
            Resource: The added resource as returned by the server.
        """
        response = requests.post(
            f"{self.url}/resource/add",
            json=resource.model_dump(mode="json"),
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def update_resource(self, resource: ResourceDataModels) -> ResourceDataModels:
        """
        Update or refresh a resource, including its children, on the server.

        Args:
            resource (ResourceDataModels): The resource to update.

        Returns:
            ResourceDataModels: The updated resource as returned by the server.
        """
        response = requests.post(
            f"{self.url}/resource/update",
            json=resource.model_dump(mode="json"),
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def get_resource(
        self,
        resource_id: str,
    ) -> ResourceDataModels:
        """
        Retrieve a resource from the server.

        Args:
            resource_id (str): The ID of the resource to retrieve.

        Returns:
            ResourceDataModels: The retrieved resource.
        """
        response = requests.get(
            f"{self.url}/resource/{resource_id}",
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def query_resource(
        self,
        resource_id: Optional[str] = None,
        resource_name: Optional[str] = None,
        parent_id: Optional[str] = None,
        resource_type: Optional[str] = None,
        resource_base_type: Optional[str] = None,
        unique: Optional[bool] = False,
        multiple: Optional[bool] = False,
    ) -> Union[ResourceDataModels, list[ResourceDataModels]]:
        """
        Query for one or more resources matching specific properties.

        Args:
            resource_id (str): The ID of the resource to retrieve.
            resource_name (str): The name of the resource to retrieve.
            parent_id (str): The ID of the parent resource.
            resource_type (str): The custom type of the resource.
            resource_base_type (str): The base type of the resource.
            unique (bool): Whether to require a unique resource or not.
            multiple (bool): Whether to return multiple resources or just the first.

        Returns:
            Resource: The retrieved resource.
        """
        payload = ResourceGetQuery(
            resource_id=resource_id,
            resource_name=resource_name,
            parent_id=parent_id,
            resource_type=resource_type,
            resource_base_type=resource_base_type,
            unique=unique,
            multiple=multiple,
        ).model_dump(mode="json")
        response = requests.post(f"{self.url}/resource/query", json=payload, timeout=10)
        response.raise_for_status()
        response_json = response.json()
        if isinstance(response_json, list):
            resources = [Resource.discriminate(resource) for resource in response_json]
            for resource in resources:
                resource.resource_url = self.url
            return resources
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def remove_resource(self, resource_id: str) -> dict[str, Any]:
        """
        Remove a resource by moving it to the history table with `removed=True`.
        """
        response = requests.delete(f"{self.url}/resource/{resource_id}", timeout=10)
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def query_history(
        self,
        resource_id: Optional[str] = None,
        version: Optional[int] = None,
        change_type: Optional[str] = None,
        removed: Optional[bool] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        limit: Optional[int] = 100,
    ) -> list[dict[str, Any]]:
        """
        Retrieve the history of a resource with flexible filters.
        """
        query = ResourceHistoryGetQuery(
            resource_id=resource_id,
            version=version,
            change_type=change_type,
            removed=removed,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
        ).model_dump(mode="json")
        response = requests.post(f"{self.url}/history/query", json=query, timeout=10)
        response.raise_for_status()

        return response.json()

    def restore_deleted_resource(self, resource_id: str) -> dict[str, Any]:
        """
        Restore a deleted resource from the history table.
        """
        response = requests.post(
            f"{self.url}/history/{resource_id}/restore", timeout=10
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def push(
        self,
        resource: Union[ResourceDataModels, str],
        child: Union[ResourceDataModels, str],
    ) -> ResourceDataModels:
        """
        Push a child resource onto a parent stack or queue.

        Args:
            resource (Union[ResourceDataModels, str]): The parent resource or its ID.
            child (Union[ResourceDataModels, str]): The child resource or its ID.

        Returns:
            ResourceDataModels: The updated parent resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        payload = PushResourceBody(
            child=child if isinstance(child, Resource) else None,
            child_id=child.resource_id if isinstance(child, Resource) else child,
        ).model_dump(mode="json")
        response = requests.post(
            f"{self.url}/resource/{resource_id}/push", json=payload, timeout=10
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def pop(
        self, resource: Union[str, ResourceDataModels]
    ) -> tuple[ResourceDataModels, ResourceDataModels]:
        """
        Pop an asset from a stack or queue resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent resource or its ID.

        Returns:
            tuple[ResourceDataModels, ResourceDataModels]: The popped asset and updated parent.
        """

        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(f"{self.url}/resource/{resource_id}/pop", timeout=10)
        response.raise_for_status()
        result = response.json()
        popped_asset = Resource.discriminate(result[0])
        update_parent = Resource.discriminate(result[1])
        popped_asset.resource_url = f"{self.url}/resource/{popped_asset.resource_id}"
        update_parent.resource_url = f"{self.url}/resource/{update_parent.resource_id}"
        return popped_asset, update_parent

    def set_child(
        self,
        resource: Union[str, ResourceDataModels],
        key: Union[str, GridIndex2D, GridIndex3D],
        child: Union[str, ResourceDataModels],
    ) -> ResourceDataModels:
        """
        Set a child resource in a parent container resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent container resource or its ID.
            key (Union[str, GridIndex2D, GridIndex3D]): The key to identify the child resource's location in the parent container.
            child (Union[str, ResourceDataModels]): The child resource or its ID.

        Returns:
            ResourceDataModels: The updated parent container resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        payload = SetChildBody(
            key=key,
            child=child,
        ).model_dump(mode="json")
        response = requests.post(
            f"{self.url}/resource/{resource_id}/child/set",
            json=payload,
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def remove_child(
        self,
        resource: Union[str, ResourceDataModels],
        key: Union[str, GridIndex2D, GridIndex3D],
    ) -> ResourceDataModels:
        """
        Remove a child resource from a parent container resource.

        Args:
            resource (Union[str, ResourceDataModels]): The parent container resource or its ID.
            key (Union[str, GridIndex2D, GridIndex3D]): The key to identify the child resource's location in the parent container.

        Returns:
            ResourceDataModels: The updated parent container resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        payload = RemoveChildBody(
            key=key,
        ).model_dump(mode="json")
        response = requests.post(
            f"{self.url}/resource/{resource_id}/child/remove",
            json=payload,
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def set_quantity(
        self, resource: Union[str, ResourceDataModels], quantity: Union[float, int]
    ) -> ResourceDataModels:
        """
        Set the quantity of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            quantity (Union[float, int]): The quantity to set.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/quantity",
            params={"quantity": quantity},
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def change_quantity_by(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Change the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to change by.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/quantity/change_by",
            params={"amount": amount},
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def increase_quantity(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Increase the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to increase by. Note that this is a magnitude, so negative and positive values will have the same effect.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/quantity/increase",
            params={"amount": amount},
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def decrease_quantity(
        self, resource: Union[str, ResourceDataModels], amount: Union[float, int]
    ) -> ResourceDataModels:
        """
        Decrease the quantity of a resource by a given amount.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            amount (Union[float, int]): The quantity to decrease by. Note that this is a magnitude, so negative and positive values will have the same effect.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/quantity/decrease",
            params={"amount": amount},
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def set_capacity(
        self, resource: Union[str, ResourceDataModels], capacity: Union[float, int]
    ) -> ResourceDataModels:
        """
        Set the capacity of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.
            capacity (Union[float, int]): The capacity to set.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/capacity",
            params={"capacity": capacity},
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def remove_capacity_limit(
        self, resource: Union[str, ResourceDataModels]
    ) -> ResourceDataModels:
        """
        Remove the capacity limit of a resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.delete(
            f"{self.url}/resource/{resource_id}/capacity", timeout=10
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def empty(self, resource: Union[str, ResourceDataModels]) -> ResourceDataModels:
        """
        Empty the contents of a container or consumable resource.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/empty",
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource

    def fill(self, resource: Union[str, ResourceDataModels]) -> ResourceDataModels:
        """
        Fill a consumable resource to capacity.

        Args:
            resource (Union[str, ResourceDataModels]): The resource or its ID.

        Returns:
            ResourceDataModels: The updated resource.
        """
        resource_id = (
            resource.resource_id if isinstance(resource, Resource) else resource
        )
        response = requests.post(
            f"{self.url}/resource/{resource_id}/fill",
            timeout=10,
        )
        response.raise_for_status()
        resource = Resource.discriminate(response.json())
        resource.resource_url = f"{self.url}/resource/{resource.resource_id}"
        return resource



"""Base node client implementation."""

from typing import Any, ClassVar

from madsci.common.types.action_types import (
    ActionRequest,
    ActionResult,
)
from madsci.common.types.admin_command_types import AdminCommandResponse
from madsci.common.types.event_types import Event
from madsci.common.types.node_types import (
    AdminCommands,
    NodeClientCapabilities,
    NodeInfo,
    NodeSetConfigResponse,
    NodeStatus,
)
from madsci.common.types.resource_types import ResourceDataModels
from pydantic import AnyUrl


class AbstractNodeClient:
    """Base Node Client, protocol agnostic, all node clients should inherit from or be based on this."""

    url_protocols: ClassVar[list[str]] = []
    """The protocol(s) to use for node URL's using this client."""

    supported_capabilities: ClassVar[NodeClientCapabilities] = NodeClientCapabilities()
    """The capabilities supported by this node client."""

    def __init__(self, url: AnyUrl) -> "AbstractNodeClient":
        """Initialize the client."""
        self.url = url

    def send_action(self, action_request: ActionRequest) -> ActionResult:
        """Perform an action on the node."""
        raise NotImplementedError("send_action not implemented by this client")

    def get_action_history(self) -> dict[str, list[ActionResult]]:
        """Get the history of a specific action, or all actions run by the node."""
        raise NotImplementedError(
            "get_action_history is not implemented by this client"
        )

    def get_action_result(self, action_id: str) -> ActionResult:
        """Get the status of an action on the node."""
        raise NotImplementedError("get_action_result is not implemented by this client")

    def get_status(self) -> NodeStatus:
        """Get the status of the node."""
        raise NotImplementedError("get_status is not implemented by this client")

    def get_state(self) -> dict[str, Any]:
        """Get the state of the node."""
        raise NotImplementedError("get_state is not implemented by this client")

    def get_info(self) -> NodeInfo:
        """Get information about the node."""
        raise NotImplementedError("get_info is not implemented by this client")

    def set_config(self, config_dict: dict[str, Any]) -> NodeSetConfigResponse:
        """Set configuration values of the node."""
        raise NotImplementedError("set_config is not implemented by this client")

    def send_admin_command(self, admin_command: AdminCommands) -> AdminCommandResponse:
        """Perform an administrative command on the node."""
        raise NotImplementedError(
            "send_admin_command is not implemented by this client"
        )

    def get_resources(self) -> dict[str, ResourceDataModels]:
        """Get the resources of the node."""
        raise NotImplementedError("get_resources is not implemented by this client")

    def get_log(self) -> dict[str, Event]:
        """Get the log of the node."""
        raise NotImplementedError("get_log is not implemented by this client")

    @classmethod
    def validate_url(cls, url: AnyUrl) -> bool:
        """check if a url matches this node type"""
        protocol = url.scheme
        return protocol in cls.url_protocols



"""REST-based node client implementation."""

import json
import tempfile
from pathlib import Path
from typing import Any, ClassVar, Optional
from zipfile import ZipFile

import requests
from madsci.client.node.abstract_node_client import (
    AbstractNodeClient,
)
from madsci.common.types.action_types import ActionRequest, ActionResult
from madsci.common.types.admin_command_types import AdminCommandResponse
from madsci.common.types.event_types import Event
from madsci.common.types.node_types import (
    AdminCommands,
    NodeClientCapabilities,
    NodeInfo,
    NodeSetConfigResponse,
    NodeStatus,
)
from madsci.common.types.resource_types.definitions import ResourceDefinition
from pydantic import AnyUrl


class RestNodeClient(AbstractNodeClient):
    """REST-based node client."""

    url_protocols: ClassVar[list[str]] = ["http", "https"]
    """The protocols supported by this client."""

    supported_capabilities: NodeClientCapabilities = NodeClientCapabilities(
        # *Supported capabilities
        get_info=True,
        get_state=True,
        get_status=True,
        send_action=True,
        get_action_result=True,
        get_action_history=True,
        action_files=True,
        send_admin_commands=True,
        set_config=True,
        get_log=True,
        # *Unsupported Capabilities
        get_resources=False,
    )

    def __init__(self, url: AnyUrl) -> "RestNodeClient":
        """Initialize the client."""
        super().__init__(url)

    def send_action(self, action_request: ActionRequest) -> ActionResult:
        """Perform an action on the node."""
        files = []
        try:
            files = [
                ("files", (file, Path(path).expanduser().open("rb")))
                for file, path in action_request.files.items()
            ]
            self.logger.log_debug(files)

            serialized_args = action_request.model_dump(mode="json")["args"]
            self.logger.log_debug(serialized_args)

            rest_response = requests.post(
                f"{self.url}/action",
                params={
                    "action_name": action_request.action_name,
                    "args": json.dumps(serialized_args),
                    "action_id": action_request.action_id,
                },
                files=files,
                timeout=10,
            )
        finally:
            # * Ensure files are closed
            for file in files:
                file[1][1].close()
        if not rest_response.ok:
            self.logger.log_error(f"{rest_response.status_code}: {rest_response.text}")
            rest_response.raise_for_status()
        return ActionResult.model_validate(rest_response.json())

    def get_action_history(
        self, action_id: Optional[str] = None
    ) -> dict[str, list[ActionResult]]:
        """Get the history of a single action performed on the node, or every action, if no action_id is specified."""
        response = requests.get(
            f"{self.url}/action", params={"action_id": action_id}, timeout=10
        )
        response.raise_for_status()
        return response.json()

    def get_action_result(self, action_id: str) -> ActionResult:
        """Get the result of an action on the node."""
        response = requests.get(
            f"{self.url}/action/{action_id}",
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        if response.files and len(response.files) == 1:
            file_key = next(iter(response.files.keys()))
            filename = response.files[file_key]
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(response.content)
                temp_path = Path(temp_file.name)
            response.files[file_key] = temp_path
        elif response.files and len(response.files) > 1:
            with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as temp_zip:
                temp_zip.write(response.content)
                temp_zip_path = Path(temp_zip.name)
            with ZipFile(temp_zip_path) as zip_file:
                for file_key in list(response.files.keys()):
                    filename = response.files[file_key]
                    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                        temp_file.write(zip_file.read(filename))
                        temp_path = Path(temp_file.name)
                    response.files[file_key] = temp_path
        return ActionResult.model_validate(response.json())

    def get_status(self) -> NodeStatus:
        """Get the status of the node."""
        response = requests.get(f"{self.url}/status", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return NodeStatus.model_validate(response.json())

    def get_state(self) -> dict[str, Any]:
        """Get the state of the node."""
        response = requests.get(f"{self.url}/state", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def get_info(self) -> NodeInfo:
        """Get information about the node."""
        response = requests.get(f"{self.url}/info", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return NodeInfo.model_validate(response.json())

    def set_config(self, new_config: dict[str, Any]) -> NodeSetConfigResponse:
        """Update configuration values of the node."""
        response = requests.post(
            f"{self.url}/config",
            json=new_config,
            timeout=60,
        )
        if not response.ok:
            response.raise_for_status()
        return NodeSetConfigResponse.model_validate(response.json())

    def send_admin_command(self, admin_command: AdminCommands) -> bool:
        """Perform an administrative command on the node."""
        response = requests.post(f"{self.url}/admin/{admin_command}", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return AdminCommandResponse.model_validate(response.json())

    def get_resources(self) -> dict[str, ResourceDefinition]:
        """Get the resources of the node."""
        raise NotImplementedError(
            "get_resources is not implemented by this client",
        )
        # TODO: Implement get_resources endpoint

    def get_log(self) -> dict[str, Event]:
        """Get the log from the node"""
        response = requests.get(f"{self.url}/log", timeout=10)
        if not response.ok:
            response.raise_for_status()
        return response.json()
